from datetime import date
from schwifty import IBAN
import requests

from ua_banktools.core import IPN


class BasePersonalClient:
    def __init__(self):
        pass


class BaseCorporateClient:
    """This class serves as a blueprint for creating new clients for corporate
    banking APIs. It includes methods usually expected from such a client."""

    USER_AGENT = "ua_banktools API Client"
    BASE_URL = ""

    def __init__(self, token: str, client_id: str = "") -> None:
        self.token = token
        self.client_id = client_id
        self.session = requests.session()

    def get_balance(self, acct: IBAN, start_date: date, end_date: date):
        pass

    def get_transactions(self, acct: IBAN, start_date: date, end_date: date):
        pass

    def create_payment(
        self,
        payer_acct: IBAN,
        recipient_acct: IBAN,
        recipient_nceo: IPN | str,
        payee_name: str,
        amount: float,
        designation: str,
        document_number: str,
    ):
        pass



import datetime as dt


class IPNValidationError(Exception):
    pass


class IPN:
    """A class representing an Individual Taxpayer Identification number for
    Ukrainian residents. Contains a 10-dicgit code, validation methods and
    properties that allow to infer the bearer's date of birth and sex."""

    def __init__(self, code: str) -> None:

        # Validation block

        # A valid IPN consists of 10 characters
        if len(code) != 10:
            raise IPNValidationError("Invalid number of digits")
        # All characters must be digits
        if not code.isdigit():
            raise IPNValidationError("Only digits are allowed in an IPN code")
        # Check control digit
        if not self.validate_control_digit(code):
            raise IPNValidationError("Control digit invalid")

        self.code = code

    def __str__(self) -> str:
        return self.code

    def __repr__(self) -> str:
        return f"IPN: {self.code}"

    @property
    def inferred_dob(self) -> dt.date:
        """The first five digits encode the bearer's date of birth: usually, it's
        the 5-digit number corresponding to the number of days elapsed from
        December 31, 1899 to the bearer's date of birth.

        However, if over 5,000 men or women happen to be born on a particular day,
        some of them may have different digits in this section of the IPN. For
        example, for January 1, 1947, the number can start with both `1` and `8`.
        """

        return dt.date(1899, 12, 31) + dt.timedelta(int(self.code[0:5]))

    @property
    def inferred_sex(self) -> str:
        """The digit before the last one encodes the bearer's sex:
        * male if the digit is odd, and
        * female if the digit is even."""

        return "female" if int(self.code[-2]) % 2 == 0 else "male"

    @staticmethod
    def validate_control_digit(code: str) -> bool:
        """The digit in the tenth (last) position serves as a checksum digit."""

        # The checksum algorithm is as follows:
        # If the code has the format `ABCDEFGHIJ`, calculate the checksum as:
        # Х = A*(-1) + B*5 + C*7 + D*9 + E*4 + F*6 + G*10 + H*5 + I*7
        d = [int(c) for c in code]
        X = (
            d[0] * -1
            + d[1] * 5
            + d[2] * 7
            + d[3] * 9
            + d[4] * 4
            + d[5] * 6
            + d[6] * 10
            + d[7] * 5
            + d[8] * 7
        )

        # The remainder from the division of the checksum by 11 will be the check
        # number: check_number = X - (11 * (X // 11)), or:
        check_number = X % 11

        # The check number may turn out to be between 0 to 10. In case the
        # remainder is 10, the `1` in the first position is dropped. The resulting
        # one digit must coincide with the last digit of the IPN.
        return check_number % 10 == d[-1]



from enum import Enum
from typing import List, Optional, Dict

from pydantic import BaseModel


# Allow arbitrary types like IBAN in requests and responses
class Config:
    arbitrary_types_allowed = True


# Enums for the Privatbank Commercial API
class Currency(Enum):
    UAH = "UAH"
    USD = "USD"
    EUR = "EUR"


class ResponseStatus(Enum):
    Success = "SUCCESS"
    Error = "ERROR"


class ResponseType(Enum):
    Balances = "balances"
    Transactions = "transactions"


class TransactionType(Enum):
    Debit = "D"
    Credit = "C"


# Data classes for the Privatbank API
class BalanceItem(BaseModel):
    acc: str
    currency: Currency
    balanceIn: float
    balanceInEq: float
    balanceOut: float
    balanceOutEq: float
    turnoverDebt: float
    turnoverDebtEq: float
    turnoverCred: float
    turnoverCredEq: float
    bgfIBrnm: str
    brnm: str
    dpd: str  # datetime: `dd.mm.yyyy hh:mm:ss`
    nameACC: str
    state: str
    atp: str
    flmn: str
    date_open_acc_reg: str  # datetime: `dd.mm.yyyy hh:mm:ss`
    date_open_acc_sys: str  # datetime: `dd.mm.yyyy hh:mm:ss`
    date_close_acc: str  # datetime: `dd.mm.yyyy hh:mm:ss`
    is_final_bal: bool


class BalanceResponse(BaseModel):
    status: ResponseStatus
    type: ResponseType
    exist_next_page: bool
    balances: List[BalanceItem]


class TransactionItem(BaseModel):
    AUT_MY_CRF: str  # My Company Code/ITIN
    AUT_MY_MFO: str  # My Bank Code
    AUT_MY_ACC: str  # My Acct. IBAN
    AUT_MY_NAM: str  # My Name
    AUT_MY_MFO_NAME: str  # My Bank
    AUT_MY_MFO_CITY: str  # My Bank City
    AUT_CNTR_CRF: str  # Counterpart's Company Code/ITIN
    AUT_CNTR_MFO: str  # Conterpart Bank Code
    AUT_CNTR_ACC: str  # Counterpart's Acct. IBAN
    AUT_CNTR_NAM: str  # Counterpart Name
    AUT_CNTR_MFO_NAME: str  # Counterpart's Bank
    AUT_CNTR_MFO_CITY: str  # Counterpart's Bank City
    CCY: Currency  # Transaction Currency
    FL_REAL: str  # "Reality" of the transaction - {r, i}
    PR_PR: str  # Transaction status - p-проводиться, t-сторнирована, r-проведена, n-забракована
    DOC_TYP: str  # Document type, e.g. "m"
    NUM_DOC: str  # Document number
    DAT_KL: str  # Client Date
    DAT_OD: str  # Remittance Date
    OSND: str  # Payment Description
    SUM: float  # Transaction Amount
    SUM_E: float  # Transaction Amount
    REF: str  # Transaction Reference
    REFN: str  # Reference number inside transaction e.g. No.1
    TIM_P: str  # Tranaction Time (most likely Ukraine time)
    DATE_TIME_DAT_OD_TIM_P: str  # Transaction DateTime
    ID: str  # Transaction ID
    TRANTYPE: TransactionType  # Transaction Type {D|C} - debit/credit
    DLR: Optional[str]  # 3rd party transaction identifier
    TECHNICAL_TRANSACTION_ID: str


class TransactionsResponse(BaseModel):
    status: ResponseStatus
    type: ResponseType
    exist_next_page: bool
    transactions: List[TransactionItem]


class PaymentCreateRequest(BaseModel):
    document_number: str
    payer_account: str
    recipient_account: str
    recipient_nceo: str
    payment_naming: str
    payment_amount: float
    payment_destination: str


class PaymentData(BaseModel):
    can_edit: str
    can_copy: Optional[str]
    document_number: str
    document_type: str
    id: str
    internal_type: str
    level_sign: Dict[str, str]
    payer_account: str
    payer_bank_name: str
    payer_name: str
    payer_nceo: str
    payment_amount: float
    payment_ccy: Currency
    payment_date_unix: str  # Unix Timestamp
    payment_destination: str
    payment_naming: str
    payment_sign: List[str]
    payment_status: str  # Enum?
    payment_status_short: str
    recipient_account: str
    recipient_ifi_text: str
    recipient_nceo: str
    service_update_utime: str
    source: str
    tabs: List[str]
    user_id: str


class PaymentCreateSuccessResponse(BaseModel):
    payment_data: PaymentData
    payment_pack_ref: str
    payment_ref: str


class ErrorResponse(BaseModel):
    status: ResponseStatus
    code: int
    message: str
    requestId: str
    serviceCode: Optional[str]



from datetime import date
import requests
from schwifty import IBAN

from ua_banktools.core import IPN
from .base import BaseCorporateClient
from .privatbank_types import *

# Privatbank API Client
class PBCorporateClient(BaseCorporateClient):
    BASE_URL = "https://acp.privatbank.ua/api/"

    def __init__(self, token: str, client_id: str) -> None:
        self.token = token
        self.client_id = client_id
        self.session = requests.session()
        self.session.headers.update(
            {
                "User-Agent": super().USER_AGENT,
                "Content-Type": "application/json;charset=utf-8",
                "id": self.client_id,
                "token": self.token,
            }
        )

    def get_balance(
        self, acct: IBAN, start_date: date, end_date: date
    ) -> BalanceResponse | ErrorResponse:
        with self.session.get(
            self.BASE_URL + "statements/balance",
            params={
                "acc": str(acct),
                "startDate": start_date.strftime("%d-%m-%Y"),
                "endDate": end_date.strftime("%d-%m-%Y"),
            },
        ) as r:
            if r.ok:
                return BalanceResponse(**r.json())
            else:
                return ErrorResponse(**r.json())

    def get_transactions(
        self, acct: IBAN, start_date: date, end_date: date
    ) -> TransactionsResponse | ErrorResponse:
        with self.session.get(
            self.BASE_URL + "statements/transactions/",
            params={
                "acc": str(acct),
                "startDate": start_date.strftime("%d-%m-%Y"),
                "endDate": end_date.strftime("%d-%m-%Y"),
            },
        ) as r:
            if r.ok:
                return TransactionsResponse(**r.json())
            else:
                return ErrorResponse(**r.json())

    def create_payment(
        self,
        payer_acct: IBAN,
        recipient_acct: IBAN,
        recipient_nceo: IPN | str,
        payee_name: str,
        amount: float,
        designation: str,
        document_number: str,
    ) -> PaymentCreateSuccessResponse | ErrorResponse:
        with self.session.post(
            self.BASE_URL + "proxy/payment/create/",
            json=PaymentCreateRequest(
                document_number=document_number,
                payer_account=str(payer_acct),
                recipient_account=str(recipient_acct),
                recipient_nceo=str(recipient_nceo),
                payment_naming=payee_name,
                payment_amount=round(amount, 2),
                payment_destination=designation,
            ).dict(),
        ) as r:
            if r.ok:
                return PaymentCreateSuccessResponse(**r.json())
            else:
                return ErrorResponse(**r.json())


"""
TODO:
* Parse dates in responses as per template (%d-%m-%Y, etc.);
* Publish to PyPI via Github Actions to be usable as a dependency
"""
